#include "afflib.h"
#include "afflib_i.h"
#include "aimage.h"

#include "imager.h"


/* checkline:
 * Looks for a name in a line in a buf and, if found,
 * make a copy of the text, put a copy of the next into the
 * AFFILE...
 */
void checkline(char *name,char *segname, char *buf,char *copy,AFFILE *af,int is_number)
{
    char *cc;
    while(buf[0] && isspace(buf[0])) buf++; // advance buf to end of spaces

    if(strncmp(buf,name,strlen(name))!=0) return;
    cc = buf+strlen(name);		// get a pointer
    while(*cc && isspace(*cc)) cc++;	// scan to end of spaces
    char *dd = index(cc,'\n');		// can we find a \n?
    if(dd) *dd = '\000';		// yes; clear it
    if(is_number){
	af_update_seg(af,segname,atoi(cc),0,0,1);   // store it in the argument...
    }
    else {
	af_update_seg(af,segname,0,cc,strlen(cc),1); // store it in the data area
	if(copy) strcpy(copy,cc);		// make a copy
    }
}




/* FreeBSD-specific ident routines */
#ifdef __FreeBSD__
#include <sys/param.h>
#include <sys/mount.h>

int freebsd_get_ata_params(const char *infile,int *controller,int *channel)
{
  int adnum;
  if(sscanf(infile,"/dev/ad%d",&adnum)==1){
      if(controller) *controller = adnum / 2;
      if(channel)    *channel    = adnum % 2;
      return 0;
  }
  return(-1);
}

#define IDENT_DEFINED
/* ident for freebsd */
void imager::ident()
{
    int controller = 0;
    int channel    = 0;

    /* Check for a SCSI drive.
     * This needs to be cleaned up; right now it assumes SCSI bus 2.
     */
    int da_dev = -1;

    if(sscanf(infile,"/dev/da%d",&da_dev)==1){
	char checkbuf[64];
	char want[64];
	char buf[256];

	/* Now do the devlist to figure out which device this is... */
	scsi_bus = -1;			// for now
	snprintf(want,sizeof(want),"da%d",da_dev);
	sprintf(checkbuf,"camcontrol devlist");
	FILE *f = popen(checkbuf,"r");
	while(!feof(f)){
	    if(fgets(buf,sizeof(buf),f)){
		if(strstr(buf,want)){	// this is the correct line
		    char *cc = strstr(buf,"at scbus");
		    if(cc){
			if(sscanf(cc,
				  "at scbus%d target %d lun %d (pass%d,da%d)",
				  &scsi_bus,&scsi_tid,&scsi_lun,
				  &scsi_pass,&da_dev)==5){
			    break;		// found it!
			}
			if(sscanf(cc,
				  "at scbus%d target %d lun %d (da%d,pass%d)",
				  &scsi_bus,&scsi_tid,&scsi_lun,
				  &da_dev,&scsi_pass)==5){
			    break;		// found it!
			}
		    }
		}
	    }
	}
	pclose(f);
	if(scsi_bus==-1) err(1,"can't find SCSI device for %s",infile);

	/* Now we need to inquiry */
	sprintf(checkbuf,"camcontrol inquiry %d:%d",scsi_bus,scsi_tid);
	f = popen(checkbuf,"r");
	char capbuf[65536];
	capbuf[0] = 0;
      
	while(!feof(f)){
	    if(fgets(buf,sizeof(buf),f)){
		strcat(capbuf,buf);
		char *cc = index(buf,'>'); // see if it should be terminated
		if(cc) *cc = 0;
		sprintf(checkbuf,"pass%d: <",scsi_pass);
		checkline(checkbuf,AF_DEVICE_MODEL,buf,device_model,af,0);
		sprintf(checkbuf,"pass%d: Serial Number ",scsi_pass);
		checkline(checkbuf,AF_DEVICE_SN,buf,serial_number,af,0);
	    }
	}
	pclose(f);
	af_update_seg(af,AF_DEVICE_CAPABILITIES,0,capbuf,strlen(capbuf),1);
	return;
    }
				     
    /* Check for an IDE drive */
    if(strncmp(infile,"/dev/ad",7)==0 &&
       freebsd_get_ata_params(infile,&controller,&channel)==0){
	char capbuf[65536];
	char buf[256];
#if __FreeBSD_version > 600000
	/* Syntax of command was changed in FreeBSD 6.0 */
	sprintf(buf,"atacontrol cap ad%d",controller*2+channel);
#else
	sprintf(buf,"atacontrol cap ata%d %d",controller,channel);
#endif
	puts(buf);
	capbuf[0] = 0;
	FILE *f = popen(buf,"r");
	while(!feof(f)){
	    if(fgets(buf,sizeof(buf),f)){
		buf[sizeof(buf)-1] = 0;	// make sure it is null-terminated
		strcat(capbuf,buf);
		checkline("device model",AF_DEVICE_MODEL,buf,device_model,af,0);
		checkline("serial number",AF_DEVICE_SN,buf,serial_number,af,0);
		checkline("firmware revision",AF_DEVICE_FIRMWARE,buf,firmware_revision,af,0);
		checkline("cylinders",AF_CYLINDERS,buf,0,af,1);
		checkline("heads",AF_HEADS,buf,0,af,1);
		checkline("sectors/track",AF_SECTORS_PER_TRACK,buf,0,af,1);
	    }
	}
	pclose(f);
	af_update_seg(af,AF_DEVICE_CAPABILITIES,0,capbuf,strlen(capbuf),1);
	return;
    }
}


/* FreeBSD ata attach commands */
void make_ata_attach_commands(char *attach,char *detach,char *master,char *slave,int dev)
{
    sprintf(attach,"atacontrol attach ata%d",dev);
    sprintf(detach,"atacontrol detach ata%d",dev);
    sprintf(master,"/dev/ad%d",dev*2);
    sprintf(slave,"/dev/ad%d",dev*2+1);
}

/* FreeBSD scsi attach commands. Actually does it and returns 0 if successful,
 * as well as the device.
 */
int  scsi_attach(char *fname,int fname_len,
		 class imager *im) // returns 0 if successful & sets fname
{
    char buf[256];
    char looking[64];

    sprintf(buf,"camcontrol rescan %d",im->scsi_bus);
    system(buf);

    sprintf(looking,"scbus%d",im->scsi_bus);	// what I'm looking for

    /* Now, see if there is a device on this scsi bus */
    FILE *f = popen("camcontrol devlist","r");
    if(!f) err(1,"camcontrol devlist");
    while(!feof(f)){
	memset(buf,0,sizeof(buf));
	if(fgets(buf,sizeof(buf),f)){
	    char *cc = strstr(buf,looking);
	    if(cc){
		int da_num;
		/* Get the target and lun.
		 * Notice that FreeBSD is inconsistent in the way that this
		 * is reported.
		 */
		if(sscanf(cc+strlen(looking)+1,
			  "target %d lun %d (pass%d,da%d)",
			  &im->scsi_tid,&im->scsi_lun,&im->scsi_pass,&da_num)==4){
		    snprintf(fname,fname_len,"/dev/da%d",da_num);
		    return 0;
		}
		if(sscanf(cc+strlen(looking)+1,
			  "target %d lun %d (da%d,pass%d)",
			  &im->scsi_tid,&im->scsi_lun,&da_num,&im->scsi_pass)==4){
		    snprintf(fname,fname_len,"/dev/da%d",da_num);
		    return 0;
		}
		err(1,"could not parse: '%s'",buf);
	    }
	}
    }
    pclose(f);
    fprintf(stderr,"No SCSI device found:\n");
    fprintf(stderr,"# camcontrol rescan all\n");
    system("camcontrol rescan all");
    return -1;				// failure
	
}

#endif





#ifdef linux
#include <dirent.h>
#define IDENT_DEFINED
void getfile(const char *dirname,const char *filename,char *copy,AFFILE *af,char *segname)
{
    char path[MAXPATHLEN];
    char buf[1024];

    /* Build the pathname we are supposed to get */
    memset(buf,0,sizeof(buf));
    strcpy(path,dirname);
    strcat(path,filename);
    if(copy) copy[0] = 0;		// by default, nothing was found
    FILE *f = fopen(path,"r");
    if(f){
	if(fgets(buf,sizeof(buf),f)){
	    char *cc = rindex(buf,'\n');
	    if(cc) *cc = '\000';	// remove trailing \n
	    af_update_seg(af,segname,0,buf,strlen(buf),1);
	    if(copy) strcat(copy,buf);	
	}
	fclose(f);
    }
}

void imager::ident()
{
    /* Check to see if infile is a USB device. If so, print things about it.
     * If the Linux /sys file system is installed, then /sys/bus/scsi/devices/.../block
     * are symlinks to the actual devices.
     * These have the same name as the /dev/<name>, minus the partition.
     */
 
    if(strncmp(infile,"/dev/",5)==0){
	char *cc;
	char sdname[MAXPATHLEN];
	memset(sdname,0,sizeof(sdname));
	strcpy(sdname,infile+5);
	/* If a partition name was provided, eliminate it */
	for(cc=sdname;*cc;cc++){
	    if(isdigit(*cc)){
		*cc = '\000';
		break;
	    }
	}
	/* Now scan the devices */
	DIR *dir = opendir("/sys/bus/scsi/devices/");
	struct dirent *dp;
	while((dp = readdir(dir))!=0){
	    if(dp->d_name[0]=='.') continue; // skip the dot names
	    char dirname[MAXPATHLEN];
	    strcpy(dirname,"/sys/bus/scsi/devices/");
	    strcat(dirname,dp->d_name);

	    char devname[MAXPATHLEN];
	    strcpy(devname,dirname);
	    strcat(devname,"/");
	    strcat(devname,"/block");

	    /* If this is a link, then stat it */
	    char path[MAXPATHLEN];
	    memset(path,0,sizeof(path));
	    if(readlink(devname,path,sizeof(path))>0){
		cc = rindex(path,'/');	// find the end of the link
		if(cc && strcmp(cc+1,sdname)==0){
		    /* Found it!
		     * Now, it turns out that dirname is also a symbolic link.
		     * Use it to find the directory where the USB information is stored.
		     */
		    char usbdir[MAXPATHLEN];
		    strcpy(usbdir,dirname);
		    strcat(usbdir,"/../../../../");
		    device_model[0] = 0;
		    serial_number[0] = 0;
		    getfile(usbdir,"manufacturer",device_model,af,AF_DEVICE_MANUFACTURER);
		    strcat(device_model," ");
		    getfile(usbdir,"product",device_model,af,AF_DEVICE_MODEL);
		    getfile(usbdir,"serial",serial_number,af,AF_DEVICE_SN);
		    getfile(usbdir,"version",0,af,AF_DEVICE_FIRMWARE);
		    return;		// we have successfully identified the device
		}
	    }
	}
    }


    if(access(infile,R_OK)==0){
	char capbuf[65536];
	char buf[256];
	sprintf(buf,"hdparm -I %s",infile);
	puts(buf);
	capbuf[0] = 0;
	FILE *f = popen(buf,"r");
	while(!feof(f)){
	    if(fgets(buf,sizeof(buf),f)){
		buf[sizeof(buf)-1] = 0;	// make sure it is null-terminated
		strcat(capbuf,buf);	// append to the buffer

		/* Now check for each of the lines */
		checkline("Model Number:",AF_DEVICE_MODEL,buf,device_model,af,0);
		checkline("Serial Number:",AF_DEVICE_SN,buf,serial_number,af,0);
		checkline("Firmware Revision:",AF_DEVICE_FIRMWARE,
			  buf,firmware_revision,af,0);
		checkline("cylinders",AF_CYLINDERS,buf,0,af,1);
		checkline("heads",AF_HEADS,buf,0,af,1);
		checkline("sectors/track",AF_SECTORS_PER_TRACK,buf,0,af,1);
	    }
	}
	pclose(f);
	af_update_seg(af,AF_DEVICE_CAPABILITIES,0,capbuf,strlen(capbuf),1);
    }
}


/*
 * Here are some parameters you may find interesting:
 *       hdparm -R 0x1700 0 0 /dev/hda   (to remove /dev/hda)
 *       ide2: BM-DMA at 0xb800-0xb807, BIOS settings: hde:DMA, hdf:pio
 */
/* Linux ata attach commands. This should work but, in my experience, it doesn't.  */
void make_ata_attach_commands(char *attach,char *detach,char *master,char *slave,int dev)
{
    system("/sbin/mknod /tmp/hda c  3 0");
    system("/sbin/mknod /tmp/hdb c  3 64");
    system("/sbin/mknod /tmp/hdc c 22 0");
    system("/sbin/mknod /tmp/hdd c 22 64");
    sprintf(attach,"hdparm -R %d",dev);
    sprintf(detach,"hdparm -U %d",dev);
    sprintf(master,"/dev/ad%d",dev*2);
    sprintf(slave,"/dev/ad%d",dev*2+1);
}

/* Linux scsi attach commands */
/* I can't figure out how to do this. */
void make_scsi_attach_commands(char *cmd_attach,char *cmd_detach, int scsi_bus)
{
    cmd_attach[0] = 0;
    cmd_detach[0] = 0;
}

int  scsi_attach(char *fname,int fname_len,class imager *) // returns 0 if successful & sets fname
{
    return -1;				// can't do it
}

#endif


/* If we do not have an ident, make it a stub... */
#ifndef IDENT_DEFINED
void imager::ident()
{
}


/* We do not have these defined either... */
void make_ata_attach_commands(char *attach,char *detach,char *master,char *slave,int dev)
{
    attach[0] = 0;
    detach[0] = 0;
    master[0] = 0;
    slave[0]  = 0;
}

int  scsi_attach(char *fname,int fname_len,class imager *) // returns 0 if successful & sets fname
{
    return -1;				// can't do it
}


#endif
